/*-
 * Copyright (C) @BABOLO  2009 Jul 14 http://www.babolo.ru/
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2009 Jul 14 http://www.babolo.ru/\n";
static const char rcsid[] = "$Id: mod_auth_pg2.c,v 1.5 2010/02/07 17:10:06 babolo Exp $";
#endif /* not lint */

/* =====================================================================
 * Copyright (c) 1996 Vidya Media Ventures, Inc. All rights reserved.
 *
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of this source code or a derived source code must
 *    retain the above copyright notice, this list of conditions and the
 *    following disclaimer. 
 *
 * 2. Redistributions of this module or a derived module in binary form
 *    must reproduce the above copyright notice, this list of conditions
 *    and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY VIDYA MEDIA VENTURES, INC. ``AS IS'' AND 
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL VIDYA MEDIA VENTURES, INC.
 * OR ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software is a contribution to and makes use of the Apache HTTP
 * server which is written, maintained and copywritten by the Apache Group.
 * See http://www.apache.org/ for more information.
 *
 * This software makes use of libpq which an interface to the PostgreSQL
 * database.  PostgreSQL is copyright (c) 1994 by the Regents of the
 * University of California.  As of this writing, more information on
 * PostgreSQL can be found at http://www.postgresSQL.org/
 *
 * PostgreSQL authentication
 * Needs libpq-fe.h and libpq.a
 *              Original source: 
 *              Adam Sussman (asussman@vidya.com) Feb, 1996
 *              Matthias Eckermann
 *              eckerman@lrz.uni-muenchen.de
 */

#include <sys/types.h>
#include <libpq-fe.h>
#include <stdlib.h>
#include <limits.h>
#include <unistd.h>
#include <errno.h>
#include <httpd.h>
#include <http_log.h>
#include <http_core.h>
#include <ap_config.h>
#include <apr_strings.h>
#include <ap_provider.h>
#include <http_config.h>
#include <http_protocol.h>
#include <babolo/BLINflag.h>

#if (MODULE_MAGIC_NUMBER_MAJOR < 20020628)
#   error  At least Apache 2.0.40 required
#endif
#if (MODULE_MAGIC_NUMBER_MAJOR < 20050708) /* Not shure XXXX */
#   define AUTH_PG2_OLDSTYLE 1
#endif

/***********************************************/
/* COPY OF httpd-2.2.11/modules/aaa/mod_auth.h */
/***********************************************/

/* Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file  mod_auth.h
 * @brief uthentication Extension Module for Apache
 *
 * @defgroup MOD_AUTH mod_auth
 * @ingroup  APACHE_MODS
 */

#ifndef APACHE_MOD_AUTH_H
#define APACHE_MOD_AUTH_H

#include "apr_pools.h"
#include "apr_hash.h"

#include "httpd.h"

#ifdef __cplusplus
extern "C" {
#endif

#define AUTHN_PROVIDER_GROUP "authn"
#define AUTHN_DEFAULT_PROVIDER "file"
    
#define AUTHZ_GROUP_NOTE "authz_group_note"
#define AUTHN_PROVIDER_NAME_NOTE "authn_provider_name"

#define AUTHN_PREFIX "AUTHENTICATE_"

typedef enum {
    AUTH_DENIED,
    AUTH_GRANTED,
    AUTH_USER_FOUND,
    AUTH_USER_NOT_FOUND,
    AUTH_GENERAL_ERROR
} authn_status;

typedef struct {
    /* Given a username and password, expected to return AUTH_GRANTED
     * if we can validate this user/password combination.
     */
    authn_status (*check_password)(request_rec *r, const char *user,
                                  const char *password);

    /* Given a user and realm, expected to return AUTH_USER_FOUND if we
     * can find a md5 hash of 'user:realm:password'
     */
    authn_status (*get_realm_hash)(request_rec *r, const char *user,
                                   const char *realm, char **rethash);
} authn_provider;

/* A linked-list of authn providers. */
typedef struct authn_provider_list authn_provider_list;

struct authn_provider_list {
    const char *provider_name;
    const authn_provider *provider;
    authn_provider_list *next;
};

typedef struct {
    /* For a given user, return a hash of all groups the user belongs to.  */
    apr_hash_t * (*get_user_groups)(request_rec *r, const char *user);
} authz_provider;

#ifdef __cplusplus
}
#endif

#endif

/***********************************************/
/* END OF httpd-2.2.11/modules/aaa/mod_auth.h */
/***********************************************/

enum classes {sq, dq, bs, sy, nl, xx};
enum states {c0, c1, cn, q0, q1, qn, k0, k1, k2, kn, s0, s1, s2, sn, ex};
enum contx {SINGLE_QU, DOUBLE_QU};

typedef struct {
    const char  *dir;
    const char  *options;
    const char  *function;
    const char  *context;
    u_int32_t    flags;
    int          authoritative;
    apr_table_t *cache_pass_table;
} auth_pg2_config_rec;

module AP_MODULE_DECLARE_DATA auth_pg2_module;
static u_int32_t flags = BLIN_VER1;

static void *
auth_pg2_create_dir_conf(apr_pool_t *p, char *d) {
    auth_pg2_config_rec *new_rec;
    int ex = 0;

    ifBLIN_QV3(flags) ap_log_perror( APLOG_MARK, APLOG_DEBUG, 0, p
                                   , "+create_dir_conf%c%s~", d ? '=' : ' ', d ? d : "NULL"
                                   );
    if  (!(new_rec = apr_pcalloc(p, sizeof(auth_pg2_config_rec)))) {
        ex = ENOMEM;
    } else {
        /* sane defaults */
        if (d) new_rec->dir = apr_pstrdup(p, d); else new_rec->dir = NULL;
        new_rec->options = NULL;
        new_rec->function = NULL;
        new_rec->context = NULL;
        new_rec->flags = flags;
        new_rec->authoritative = 1;
    }
    if  (ex) new_rec = NULL;
    ifBLIN_QV3(flags) ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, p, "-create_dir_conf %d", ex);
    return(new_rec);
}

static size_t
auth_pg2_ws(request_rec *r, enum contx q, const char *from, char **to, const char *p) {
    size_t n;
    const u_char *in;
    enum states state;
    u_int32_t control = 0;
    const enum classes class[256] =
    {nl,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,dq,sy ,sy,sy,sy,sq ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy

    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,bs,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy

    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy

    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    };
#define C_a2 0x40000000 /* ++2                 */
#define C_a1 0x20000000 /* ++1                 */
#define C_bs 0x10000000 /* \\                  */
#define C_sq  0x8000000 /* copy '              */
#define C_dq  0x4000000 /* copy "              */
#define C_sk  0x2000000 /* copy '              */
#define C_dk  0x1000000 /* copy "              */
#define C_cp   0x800000 /* copy symbol         */
#define C_wp   0x400000 /* switch to p         */
#define C_u0   0x200000 /* in++                */
#define C_state    0xFF /*   */
#define C_bc C_bs|C_cp
    const u_int32_t automa[ex][xx] =
    /*               sq,                dq,                bs,           sy,                nl */
    {{     C_a2|C_u0|c0,      C_a1|C_u0|c0,      C_a2|C_u0|c0, C_a1|C_u0|c0,      C_a1|C_wp|c1}/*c0*/
    ,{     C_a1|C_u0|c1,      C_a1|C_u0|c1,      C_a1|C_u0|c1, C_a1|C_u0|c1,                ex}/*c1*/
    ,{     C_a1|C_u0|cn,      C_a1|C_u0|cn,      C_a1|C_u0|cn, C_a1|C_u0|cn,           C_wp|c1}/*cn*/
    ,{     C_a1|C_u0|q0,      C_a2|C_u0|q0,      C_a2|C_u0|q0, C_a1|C_u0|q0,      C_a1|C_wp|q1}/*q0*/
    ,{     C_a1|C_u0|q1,      C_a1|C_u0|q1,      C_a1|C_u0|q1, C_a1|C_u0|q1,                ex}/*q1*/
    ,{     C_a1|C_u0|qn,      C_a1|C_u0|qn,      C_a1|C_u0|qn, C_a1|C_u0|qn,           C_wp|q1}/*qn*/
    ,{C_sq|C_bs|C_cp|k1, C_sq     |C_cp|k1, C_sq|C_bs|C_cp|k1, C_sq|C_cp|k1, C_sq|C_sk|C_wp|k2}/*k0*/
    ,{     C_bs|C_cp|k1,           C_cp|k1,      C_bs|C_cp|k1,      C_cp|k1,      C_sk|C_wp|k2}/*k1*/
    ,{          C_cp|k2,           C_cp|k2,           C_cp|k2,      C_cp|k2,                ex}/*k2*/
    ,{          C_cp|kn,           C_cp|kn,           C_cp|kn,      C_cp|kn,           C_wp|k2}/*kn*/
    ,{C_dq     |C_cp|s1, C_dq|C_bs|C_cp|s1, C_dq|C_bs|C_cp|s1, C_dq|C_cp|s1, C_dq|C_dk|C_wp|s2}/*s0*/
    ,{          C_cp|s1,      C_bs|C_cp|s1,      C_bs|C_cp|s1,      C_cp|s1,      C_dk|C_wp|s2}/*s1*/
    ,{          C_cp|s2,           C_cp|s2,           C_cp|s2,      C_cp|s2,                ex}/*s2*/
    ,{          C_cp|sn,           C_cp|sn,           C_cp|sn,      C_cp|sn,           C_wp|s2}/*sn*/
    };

    if  (q == SINGLE_QU) {
        if  (from == NULL) {
            from = "NULL";
            state = (to) ? kn : cn;
            n = 0;
        } else {
            state = (to) ? k0 : c0;
            n = 1;
        }
    } else if (q == DOUBLE_QU) {
        if  (from == NULL) {
            from = "NULL";
            state = (to) ? sn : qn;
            n = 0;
        } else {
            state = (to) ? s0 : q0;
            n = 1;
        }
    } else {                                                /* Low priority to display - impossible */
        ifBLIN_QV1(((auth_pg2_config_rec *)ap_get_module_config(r->per_dir_config, &auth_pg2_module))->flags)
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "ws Internal error #1");
        return(0);
    };
    for (in = (const u_char *)from; state < ex;) {
        control = automa[state][class[*in]];
        if  (control & C_a1) n += 1;
        if  (control & C_a2) n += 2;
        if  (control & C_bs) *(*to)++ = '\\';
        if  (control & C_sq) *(*to)++ = '\'';
        if  (control & C_dq) *(*to)++ = '\"';
        if  (control & C_sk) *(*to)++ = '\'';
        if  (control & C_dk) *(*to)++ = '\"';
        if  (control & C_cp) *(*to)++ = *in++;
        if  (control & C_wp) in = (const u_char *)p;
        if  (control & C_u0) in++;
        state = control & C_state;
    }
    return(n);
};

static authn_status
auth_pg2_check_password(request_rec *r, const char *user, const char *sent_pw) {
    char *query, *q, *val, *creds;
    auth_pg2_config_rec *sec;
    PGconn *pg_conn = NULL;
    PGresult *pg_result;
    authn_status res;
    size_t sums;
    int ex = 0;

    sec = (auth_pg2_config_rec *)ap_get_module_config(r->per_dir_config, &auth_pg2_module);
    ifBLIN_QV3(sec->flags)
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "+check_password=%s:%s~", user, sent_pw);
    if  (!sec->function) {
        ifBLIN_QV2(sec->flags) ap_log_rerror( APLOG_MARK, APLOG_WARNING, 0, r
                                            , "check_password: missing Auth_PG2_function"
                                            );
        res = AUTH_USER_NOT_FOUND;
        ex = 1;
    }
    sums = 67;                                                             /* include trailing '\0' */
    if  (!ex) {
        sums += auth_pg2_ws(r, DOUBLE_QU, sec->function           , NULL, "(");
        if  (sec->context) sums += auth_pg2_ws(r, SINGLE_QU, sec->context, NULL, ", ");
        sums += auth_pg2_ws(r, SINGLE_QU, user                    , NULL, ", ");
        sums += auth_pg2_ws(r, SINGLE_QU, sent_pw                 , NULL, ", ");
        sums += auth_pg2_ws(r, SINGLE_QU, r->connection->remote_ip, NULL, ", ");
        sums += auth_pg2_ws(r, SINGLE_QU, r->uri                  , NULL, ", ");
        sums += auth_pg2_ws(r, SINGLE_QU, r->args                 , NULL, ", ");
        sums += auth_pg2_ws(r, SINGLE_QU, r->protocol             , NULL, ")");
        if  ((query = apr_palloc(r->pool, sums)) == NULL) {
            ifBLIN_QV1(sec->flags)
                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "check_password: No mem for query string");
            res = AUTH_GENERAL_ERROR;
            ex = 1;
    }   };
    if  (!ex) {
        strcpy(query, "SELECT ");
        strcat(query, sec->function);
        strcat(query, "(");
        q = index(query, '\0');
        if  (sec->context) auth_pg2_ws(r, SINGLE_QU, sec->context, &q, "::name, ");
        auth_pg2_ws(r, SINGLE_QU, user                    , &q, "::name, ");
        auth_pg2_ws(r, SINGLE_QU, sent_pw                 , &q, "::text, ");
        auth_pg2_ws(r, SINGLE_QU, r->connection->remote_ip, &q, "::inet, ");
        auth_pg2_ws(r, SINGLE_QU, r->uri                  , &q, "::text, ");
        auth_pg2_ws(r, SINGLE_QU, r->args                 , &q, "::text, ");
        auth_pg2_ws(r, SINGLE_QU, r->protocol             , &q, "::text)");
        *q = '\0';
        if  (pg_conn && PQstatus(pg_conn) != CONNECTION_OK) {
            PQfinish(pg_conn);
            pg_conn = NULL;
    }   }
    if  (!ex && !pg_conn) {
        pg_conn = PQconnectdb(sec->options ? sec->options : "");
        if  (!pg_conn) {
            ifBLIN_QV1(sec->flags)
                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "check_password: No connection to DBMS");
            res = AUTH_GENERAL_ERROR;
            ex = 1;
        }else if (PQstatus(pg_conn) != CONNECTION_OK) {
            ifBLIN_QV1(sec->flags) ap_log_rerror( APLOG_MARK, APLOG_ERR, 0, r
                                                , "check_password: PQ #3 %s"
                                                , PQerrorMessage(pg_conn)
                                                );
            res = AUTH_GENERAL_ERROR;
            ex = 1;
    }   }
    creds = NULL;
    if  (!ex) {
        pg_result = PQexec(pg_conn, query);
        if  (!pg_result) {
            ifBLIN_QV1(sec->flags) ap_log_rerror( APLOG_MARK, APLOG_ERR, 0, r
                                                , "check_password: PQ #1 %s"
                                                , PQerrorMessage(pg_conn)
                                                );
            ex = 1;
    }   }
    if  (!ex && (PQresultStatus(pg_result) != PGRES_TUPLES_OK)) {
        ifBLIN_QV1(sec->flags) ap_log_rerror( APLOG_MARK, APLOG_ERR, 0, r
                                            , "check_password: PQ #2 %s"
                                            , PQerrorMessage(pg_conn)
                                            );
        PQclear(pg_result);
        ex = 1;
    }
    if  (!ex && (PQntuples(pg_result) >= 1) && !PQgetisnull(pg_result, 0, 0)) {
        val = PQgetvalue(pg_result, 0, 0);
        if  (!(creds = (char *)apr_palloc(r->pool, PQgetlength(pg_result, 0, 0) + 1))) {
            ifBLIN_QV1(sec->flags)
                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "check_password: No mem for query result");
            PQclear(pg_result);
            ex = 1;
        }
        strcpy(creds, val);
    }
    /* ignore errors here ! */
    if  (!ex) PQclear(pg_result);
    if  (pg_conn) {
        PQfinish(pg_conn);
        pg_conn = NULL;
    }
    if  (ex || !creds || !creds[0]) {
        if  (sec->authoritative) {
            /* force error and access denied */
            ifBLIN_QV1(sec->flags)
                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "check_password: User %s failed", user);
            ap_note_basic_auth_failure(r);
            res = AUTH_DENIED;
        } else {
            /* allow fall through to another module */
            ifBLIN_QV2(sec->flags)
                ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r, "check_password: User %s fall", user);
            res = AUTH_USER_NOT_FOUND;
        }
        ex = 1;
    }
    if  (!ex) apr_table_setn(r->subprocess_env, "CREDENTIALS", creds);
    ifBLIN_QV3(sec->flags)
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "-check_password %d=%s~", res, creds);
    return(!ex ? AUTH_GRANTED : res);
}

#ifdef AUTH_PG2_OLDSTYLE

static int
auth_pg2_check_user_id(request_rec *r) {
    auth_pg2_config_rec *sec;
    const char *sent_pw;
    authn_status res;
    int ex;

    sec = (auth_pg2_config_rec *)ap_get_module_config(r->per_dir_config, &auth_pg2_module);
    ifBLIN_QV3(sec->flags) ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "+check_user_id");
    if  (!(ex = ap_get_basic_auth_pw(r, &sent_pw))) {
        /* Dont know about int <=> authn_status compatibility so recode */
        res = auth_pg2_check_password(r, r->user, sent_pw);
        switch (res) {
        case AUTH_USER_NOT_FOUND: ex = DECLINED         ; break;
        case AUTH_DENIED        : ex = HTTP_UNAUTHORIZED; break;
        case AUTH_GRANTED       : ex = OK               ; break;
        case AUTH_GENERAL_ERROR :
        default                 : ex = HTTP_INTERNAL_SERVER_ERROR;
    }   }
    ifBLIN_QV3(sec->flags) ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "-check_user_id %d", ex);
    return(ex);
}

#endif

/* Checking ID */
static int
auth_pg2_auth_checker(request_rec *r) {
    apr_array_header_t *reqs_arr = (apr_array_header_t *)ap_requires(r);
    require_line *reqs = reqs_arr ? (require_line *)reqs_arr->elts : NULL;
    int m = r->method_number;
    auth_pg2_config_rec *sec;
    char *user = r->user;
    const char *t;
    int x, res, ex;
    char *w;

    sec = (auth_pg2_config_rec *)ap_get_module_config(r->per_dir_config, &auth_pg2_module);
    ifBLIN_QV3(sec->flags) ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "+auth_checker");
    res = DECLINED;
    ex = 0;
    /* if we cannot do it; leave it to some other guy */
    if  (!sec->function) ex = 1;
    if  (!ex && !reqs_arr) {
        if  (sec->authoritative) {
            /* force error and access denied */
            ifBLIN_QV1(sec->flags) ap_log_rerror( APLOG_MARK, APLOG_ERR, 0, r
                                                , "+auth_checker: No rules specified #1 for user %s"
                                                , user
                                                );
            ap_note_basic_auth_failure(r);
            res = HTTP_UNAUTHORIZED;
        } else {
            res = DECLINED;
        }
        ex = 1;
    }
    for (x = 0; !ex && (x < reqs_arr->nelts); x++) {
        if  (!(reqs[x].method_mask & (1 << m))) continue;
        t = reqs[x].requirement;
        w = ap_getword(r->pool, &t, ' ');
        if  (!strcmp(w, "valid-user")) {
            res = OK;
            ex = 1;
            break;
        }
        if  (!strcmp(w, "user")) {
            while (t[0]) {
                w = ap_getword_conf(r->pool, &t);
                if  (!strcmp(user, w)) {
                    res = OK;
                    ex = 1;
                    break;
            }   }
            if  (sec->authoritative) {
                /* force error and access denied */
                ifBLIN_QV1(sec->flags) ap_log_rerror( APLOG_MARK, APLOG_ERR, 0, r
                                                    , "auth_checker: No rules specified #2 for user %s"
                                                    , user
                                                    );
                ap_note_basic_auth_failure(r);
                res = HTTP_UNAUTHORIZED;
                ex = 1;
                break;
    }   }   }
    ifBLIN_QV3(sec->flags) ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "-auth_checker %d", res);
    return(res);
}

#if 0
static authn_status
auth_pg2_get_realm_hash(request_rec *r, const char *user, const char *realm, char **rethash) {
    auth_pg2_config_rec *sec;
    authn_status res;

    sec = (auth_pg2_config_rec *)ap_get_module_config(r->per_dir_config, &auth_pg2_module);
    ifBLIN_QV3(sec->flags) ap_log_rerror( APLOG_MARK, APLOG_DEBUG, 0, r
                                        , "+get_realm_hash=%s:%s~", user, realm
                                        );
    *rethash = "0123456789abcdef0123456789abcdef";
    res = AUTH_USER_FOUND;
    ifBLIN_QV3(sec->flags) ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "-get_realm_hash %d", res);
    return(res);
}
#endif

static int
auth_pg2_post_config(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s) {
    ifBLIN_QV3(flags) ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, p, "+post_config");
    ap_add_version_component(p, "mod_auth_pg2-" VERS);
    ifBLIN_QV3(flags) ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, p, "-post_config %d", OK);
    return(OK);
}

static const authn_provider auth_pg2_provider =
{ &auth_pg2_check_password
, NULL /* &auth_pg2_get_realm_hash */
};

static void
auth_pg2_register_hooks(apr_pool_t *p) {
    ifBLIN_QV3(flags) ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, p, "+register_hooks");
    ap_hook_post_config(auth_pg2_post_config, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_auth_checker(auth_pg2_auth_checker, NULL, NULL, APR_HOOK_MIDDLE);
#ifdef AUTH_PG2_OLDSTYLE
    ap_hook_check_user_id(auth_pg2_check_user_id, NULL, NULL, APR_HOOK_MIDDLE);
#else
    ap_register_provider(p, AUTHN_PROVIDER_GROUP, "pg2", "0", &auth_pg2_provider);
#endif
    ifBLIN_QV3(flags) ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, p, "-register_hooks");
};

static const char *
auth_pg2_setblin(cmd_parms *cmd, void *mconfig, const char *level) {
    auth_pg2_config_rec *sec = mconfig;
    u_int32_t b;
    int32_t l;

    l = strtol(level, NULL, 0);
    if  (l > 8) l = 8;
    if  (l < 0) l = 0;
    for (b = 0; l > 0; l--) BLIN_VERBOSE(b);
    sec->flags = b;
    return(NULL);
}

static const command_rec auth_pg2_cmds[] =
{   AP_INIT_TAKE1( "Auth_PG2_options"
                 , ap_set_string_slot
                 , (void*)APR_OFFSETOF(auth_pg2_config_rec, options)
                 , OR_AUTHCFG
                 , "PQconnectdb() parameter."
                 )
,   AP_INIT_TAKE1( "Auth_PG2_function"
                 , ap_set_string_slot
                 , (void*)APR_OFFSETOF(auth_pg2_config_rec, function)
                 , OR_AUTHCFG
                 , "the name of the function to handle username/password tuples."
                 )
,   AP_INIT_TAKE1( "Auth_PG2_context"
                 , ap_set_string_slot
                 , (void*)APR_OFFSETOF(auth_pg2_config_rec, context)
                 , OR_AUTHCFG
                 , "the name of the context."
                 )
,   AP_INIT_TAKE1( "Auth_PG2_verbose"
                 , auth_pg2_setblin
                 , NULL
                 , OR_AUTHCFG
                 , "verbose level."
                 )
,   AP_INIT_FLAG( "Auth_PG2_authoritative"
                , ap_set_flag_slot
                , (void*)APR_OFFSETOF(auth_pg2_config_rec, authoritative)
                , OR_AUTHCFG
                , "'on' or 'off'"
                )
,   {NULL}
};

module AP_MODULE_DECLARE_DATA auth_pg2_module =
{   STANDARD20_MODULE_STUFF
,   auth_pg2_create_dir_conf       /* dir config creater                    */
,   NULL                           /* dir merger --- default is to override */
,   NULL                           /* server config                         */
,   NULL                           /* merge server config                   */
,   auth_pg2_cmds                  /* command table                         */
,   auth_pg2_register_hooks        /* Apache2 register hooks                */
};
